/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: bytecode.c,v 1.9 2000/08/05 22:25:17 malekith Exp $ */

#include "vm.h"
#include "parser.h"

/* ALIGN macro is used also below, undef it if your CPU can handle
 * unaligned access 
 * you may also define it to value of align your CPU needs,
 * sizeof(void*) could be bigger then it's needed, so it
 * can enlarge memory requirements
 * note that ALIGN has to be power of 2 !
 */
#ifdef __i386__
#undef ALIGN
#else
#define ALIGN sizeof(void*)
#endif

/* structure:
 * u16 head
 *	head & 63 == id
 *	(head >> 6) == sel
 * u8 file_idx
 * u16 line 
 * (sel & SEL_Z_PTRS) ? (
 *   (sel & SEL_NEXT) ? u8 next_offset : ()
 *   (sel & SEL_LEFT) ? u8 left_offset : ()
 *   (sel & SEL_RIGHT) ? u8 right_offset : ()
 * ) : (
 *   (sel & SEL_NEXT) ? u16 next_offset : ()
 *   (sel & SEL_LEFT) ? u16 left_offset : ()
 *   (sel & SEL_RIGHT) ? u16 right_offset : ()
 * )
 * (sel & SEL_MASK) ? ....
 */

/* if first four bytes of a byte code equals to this,
 * then it is assumed to be start of byte code, with std header:
 * u32 magic, offset_of_first, offset_of_first_ptr, number_of_ptrs;
 * ptrs, are ptrs to regexps, which need to be freed when
 * bc is freed */
byte *bc_magic = "\xff\xff\xff\xff";

enum {
SEL_MASK 	= 0x000f << 6,
SEL_INT_4 	= 0x0001 << 6,	/* 4 byte signed int */
SEL_INT_2	= 0x0002 << 6,	/* 2 byte signed int */
SEL_INT_1	= 0x0003 << 6,  /* 1 byte signed int */
SEL_INT_U1	= 0x0004 << 6,  /* 1 byte unsigned int | 0x100 */
SEL_INT_1_100	= 0x0005 << 6,  /* 1 byte unsigned int | 0x100 */
SEL_CHAR	= 0x0006 << 6,  /* nul terminated string */
SEL_CHAR_R1	= 0x0007 << 6,  /* 1 byte unsigned offset to string */
SEL_CHAR_R2	= 0x0008 << 6,  /* 2 byte unsigned offset to string */
SEL_PTR		= 0x0009 << 6,  /* 4 byte (void*) */
SEL_TOP_1	= 0x000a << 6,  /* 1 byte offset to (vm_token) */
SEL_TOP_2	= 0x000b << 6,  /* 2 byte offset to (vm_token) */
SEL_RE_1	= 0x000c << 6,	/* 1 byte offset to compiled regexp */
SEL_RE_2	= 0x000d << 6,  /* 2 byte offset to compiled regexp */
SEL_UNUSED_E	= 0x000e << 6,
SEL_UNUSED_F	= 0x000f << 6,

SEL_NEXT	= 0x0010 << 6,	/* 1 | 2 byte offset to (vm_token) */
SEL_LEFT 	= 0x0020 << 6,  /* " */
SEL_RIGHT 	= 0x0040 << 6,  /* " */
SEL_Z_PTRS	= 0x0080 << 6,  /* if set 3 above offsets are 1 byte */

SEL_LINE_17	= 0x0100 << 6,	/* 17th bit of line */
SEL_LINE_18	= 0x0200 << 6,  /* 18th bit of line */
};

struct output {
	int pos, left;
	byte *start;
	int stack;

	hash *lookup;
};

struct entry {
	int id, file_idx;
	int line;
	int xnext, xleft, xright, xtop;
	int xint;
	char *xchar;
	void *xptr;
	int xre;
};

static inline void lookup_add(struct output *o)
{
	var *v;
	
	if (o->lookup == 0) {
		o->lookup = hash_new(0);
	}
	v = new_iv(o->start + o->pos, o->pos);
	hash_add(o->lookup, v);
}

static inline int lookup_string(struct output *o, const char *string)
{
	var *v;
	
	if (o->lookup == 0)
		return 0;
	v = hash_find_ex(o->lookup, string);
	
	if (v)
		return v->var_int;
	else
		return 0;
}


static void alloc(struct output *o, int size)
{
	if (o->left < size) {
		if (o->stack) {
			byte *tmp = o->start;
			
			o->start = xmalloc(o->pos + 512 + size);
			memcpy(o->start, tmp, o->pos);
			o->stack = 0;
		} else
			o->start = xrealloc(o->start, 
						o->pos + 512 + size);
		o->left = 510;
	} else
		o->left -= size;
}

static int put_aligned(struct output *o, void *src, int src_len)
{
	int r;
	
#ifdef ALIGN
	r = ALIGN - (o->pos & (ALIGN - 1));
	if (r == ALIGN)
		r = 0;
	alloc(o, src_len + r);
	o->pos += r;
#else
	alloc(o, src_len);
#endif

	r = o->pos;
	if (src_len)
		memcpy(o->start + o->pos, src, src_len);
	o->pos += src_len;

	return r;
}

#define PUT(data, type)							\
	do { 								\
		type __tmp;						\
									\
		__tmp = (type)(data);					\
		memmove(o->start + o->pos, &__tmp, sizeof(type)); 	\
		o->pos += sizeof(type); 				\
	} while (0)

static inline int compute_entry_size(struct entry *e, struct output *o,
				     u16 * ssel, u32 * ttmp)
{
	u16 sel;
	u32 tmp;
	int can_compress = 1, size = 0, k = 0;
	
	tmp = *ttmp;
	sel = *ssel;

	if ((sel = (e->id & 63)) != e->id)
		abort();

	sel |= (e->line & 0x30000) >> 2;
#define X(xnext, se) 					\
	if (e->xnext) {					\
		if ((o->pos - e->xnext) & 0xffff0000)	\
			abort();			\
		if ((o->pos - e->xnext) & 0xff00)	\
			can_compress=0;			\
		size++;					\
		sel |= se;				\
	} 						\
	(void) 0

	X(xnext, SEL_NEXT);
	X(xleft, SEL_LEFT);
	X(xright, SEL_RIGHT);

	if (can_compress)
		sel |= SEL_Z_PTRS;
	else
		size *= 2;

	/* u16 sel, u8 fidx, u16 line == bytes */
	size += 5;

	if (e->xtop) {
		if ((tmp = (u32) (o->pos - e->xtop)) & 0xffff0000)
		    abort();
		if (tmp & 0xff00) {
			size += 2;
			sel |= SEL_TOP_2;
		} else {
			sel |= SEL_TOP_1;
			size++;
		}
		k++;
	}

	if (e->xint) {
		size++;
		tmp = e->xint;
		if ((e->xint & 0xffffff80) == 0
		    || (e->xint & 0xffffff80) == 0xffffff80) 
			sel |= SEL_INT_1;
		else if ((e->xint & 0xffffff00) == 0)
			sel |= SEL_INT_U1;
		else if ((e->xint & 0xfffffe00) == 0) 
			sel |= SEL_INT_1_100;
		else if ((e->xint & 0xffff8000) == 0
			   || (e->xint & 0xffff8000) == 0xffff8000) {
			sel |= SEL_INT_2;
			size++;
		} else {
			sel |= SEL_INT_4;
			size += 3;
		}
		k++;
	}

	if (e->xchar) {
		if ((tmp = lookup_string(o, e->xchar)))
			tmp = o->pos - tmp;
		size++;
		if (tmp == 0 || (tmp & 0xffff0000)
		    || e->xchar[0] == 0
		    || (e->xchar[1] == 0 && (tmp & 0xff00))) {
			size += strlen(e->xchar);
			sel |= SEL_CHAR;
		} else if (tmp & 0xff00) {
			sel |= SEL_CHAR_R2;
			size++;
		} else
			sel |= SEL_CHAR_R1;

		k++;
	}

	if (e->xptr) {
		size += sizeof(void *);
		sel |= SEL_PTR;
		k++;
	}

	if (e->xre) {
		tmp = o->pos - e->xre;
		if (tmp & 0xffff0000)
			abort();
		if (tmp & 0xff00) {
			size += 2;
			sel |= SEL_RE_2;
		} else {
			sel |= SEL_RE_1;
			size++;
		}
		k++;
	}

	if (k > 1)
		abort();
	*ttmp = tmp;
	*ssel = sel;
	return size;
}

static inline void write_entry(struct entry *e, struct output *o, 
				u16 sel, u32 tmp)
{
	int off;
	
	off = o->pos;
	
	PUT(sel, u16);
	PUT(e->file_idx, u8);
	PUT(e->line & 0xffff, u16);

#define drop(x) 					\
	if (e->x) { 					\
		if (sel & SEL_Z_PTRS)			\
			PUT(off - e->x, u8);		\
		else		 			\
			PUT(off - e->x, u16);		\
	} (void)0

	drop(xnext);
	drop(xleft);
	drop(xright);

	switch (sel & SEL_MASK) {
	case SEL_PTR:
		PUT(e->xptr, void *);
		break;
	case SEL_CHAR:
		strcpy(o->start + o->pos, e->xchar);
		lookup_add(o);
		o->pos += strlen(e->xchar) + 1;
		break;
	case SEL_INT_4:
		PUT(tmp, u32);
		break;
	case SEL_CHAR_R2:
	case SEL_INT_2:
	case SEL_TOP_2:
	case SEL_RE_2:
		PUT(tmp & 0xffff, u16);
		break;
	case SEL_CHAR_R1:
	case SEL_INT_U1:
	case SEL_INT_1:
	case SEL_INT_1_100:
	case SEL_TOP_1:
	case SEL_RE_1:
		PUT(tmp & 0xff, u8);
		break;
	}
} 

static void pack_entry(struct entry *e, struct output *o)
{
	u16 sel = 0;
	u32 tmp = 0;

	alloc(o, compute_entry_size(e, o, &sel, &tmp));
	write_entry(e, o, sel, tmp);
}

#ifdef ALIGN
# ifdef WORDS_BIGENDIAN
#  define GET16(a)					\
	do { 						\
		a = ((int)p[1] << 16) + (int)p[0]; 	\
		p += 2; 				\
	} while (0)
# else
#  define GET16(a)					\
	do { 						\
		a = ((int)p[0] << 16) + (int)p[1]; 	\
		p += 2; 				\
	} while (0)
# endif
#else
/* this wouldn't work on e.g. sparc, but works perfectly on i386 */
# define GET16(a) 					\
	do { 						\
		a = *((u16*)p); 			\
		p += 2; 				\
	} while (0)
#endif

#define GET(data, type)							\
	do { 								\
		type __tmp;						\
									\
		memmove(&__tmp, p, sizeof(type)); 			\
		data = __tmp;						\
	} while (0)

void unpack(vm_token *t, byte *p)
{
	u16 sel;
	int tmp;
	byte *start;

	/* if it's first one, move p to begining of the first */
	if (memcmp(p, bc_magic, 4) == 0) 
		p += ((u32*)p)[1];
		
	start = p;
	
	memset(t, 0, sizeof(*t));

	GET16(sel);
	t->id = sel & 63;
	t->file_idx = *p++;
	GET16(t->line);

	t->line |= ((sel << 2) & 0x30000);

	if (sel & SEL_Z_PTRS) {
		if (sel & SEL_NEXT)
			t->next = start - *p++;
		if (sel & SEL_LEFT)
			t->left = start - *p++;
		if (sel & SEL_RIGHT)
			t->right = start - *p++;
	} else {
		u16 d;

		if (sel & SEL_NEXT) {
			GET16(d);
			t->next = start - d;
		}
		if (sel & SEL_LEFT) {
			GET16(d);
			t->left = start - d;
		}
		if (sel & SEL_RIGHT) {
			GET16(d);
			t->right = start - d;
		}
	}
	
	switch (sel & SEL_MASK) {
	case 0:
		break;
	case SEL_TOP_1:
		t->u.top = start - *p;
		break;
	case SEL_TOP_2:
		GET16(tmp);
		t->u.top = start - tmp;
		break;
	case SEL_INT_4:
		GET(t->u.number, i32);
		break;
	case SEL_INT_2:
		GET(t->u.number, i16);
		break;
	case SEL_INT_U1:
		GET(t->u.number, u8);
		break;
	case SEL_INT_1:
		GET(t->u.number, i8);
		break;
	case SEL_INT_1_100:
		GET(t->u.number, u8);
		t->u.number |= 0x100;
		break;
	case SEL_CHAR:
		t->u.name = p;
		break;
	case SEL_CHAR_R1:
		t->u.name = start - *p;
		break;
	case SEL_CHAR_R2:
		GET16(tmp);
		t->u.name = start - tmp;
		break;
	case SEL_PTR:
		GET(t->data_ptr, void*);
		break;
	case SEL_RE_1:
		t->data_ptr = start - *p;
		break;
	case SEL_RE_2:
		GET16(tmp);
		t->data_ptr = start - tmp;
		break;
	default:
		abort();
	}
//printf("%p: id: %d (%s) (%p %p %p %p)\n",start,t->id,id_name(t->id),t->next,t->left,t->right,t->u.top);
}

static int pack_token(token *t, struct output *o)
{
	struct entry e;
	int off;

	if (t == 0)
		return 0;
		
	memset(&e, 0, sizeof(e));
	
	e.xint = t->u.number;
	e.xchar = t->u.name;
	
	e.file_idx = t->file_idx;
	e.line = t->line;
	e.xnext = pack_token(t->next, o);
	e.xright = pack_token(t->right, o);
	e.xleft = pack_token(t->left, o);
	e.xtop = pack_token(t->u.top, o);
	t->u.top = t->next = t->left = t->right = 0;
	
	if ((e.id = t->id) == ID_REGEXP) {
		e.xre = put_aligned(o, t->data_ptr, t->data_len);
		xfree(t->data_ptr);
		t->data_ptr = 0;
		t->data_len = 0;
	}
	
	if (e.id == ID_ERROR)
		abort();
		
	off = o->pos;
	pack_entry(&e, o);
	
	kill_token(t);
	
//printf("packed at %d\n", off);
	return off;
}

/* this function consumes a lot of stack, but thanx to this
 * it doesn't have to extensivly use xmalloc for small ( under 8k ;)
 * functions */
byte *pack(token *t, int more, byte **more_ptr, int *total_size)
{
	struct output out;
	u32 start;
	static int total;
	byte stack[8030];
	byte *r;

	if (t == 0)
		return 0;
		
	out.start = stack;
	out.stack = 1;
	out.pos = 8;
	out.left = 8000;
	out.lookup = 0;

	start = pack_token(t, &out);

	/* align out.pos if there's request for some memory */
	if (more)
		put_aligned(&out, 0, 0);
	
	total += out.pos + more;
	
	if(getenv("DUMP_PACK"))
		printf("packed: %d @ %d + %d bytes (%d)\n",
		       out.pos, start, more, total);
	
	r = (byte*)xmalloc(out.pos + more);
	memcpy(r, out.start, out.pos);
	
	if (more_ptr)
		*more_ptr = r + out.pos;
	if (total_size)
		*total_size = out.pos + more;
		
	memcpy(r, bc_magic, 4);
	memcpy(r + 4, &start, 4);
	
	if (!out.stack)
		xfree(out.start);
	if (out.lookup)
		hash_kill_one(out.lookup);
		
	return r;
}

void free_bc(byte * code)
{
	if (code == 0)
		return;
		
	/* check for bad magic */
	if (memcmp(code, bc_magic, 4))
		abort();
		
	xfree(code);
}
