/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: flow.c,v 1.14 2000/08/04 21:28:14 malekith Exp $ */

#include "vm.h"
#include "proc.h"

void setup_frame(frame *ret, frame *parent)
{
	ret->hashes = 0;
	ret->stack = hash_new(0);
	ret->break_lev = 0;
	ret->continue_lev = 0;
	ret->return_lev = 0;

	ret->regexp = 0;
	
	/* global frame has itself as a global */
	ret->global = parent ? parent->global : ret;
	ret->parent = parent;
	ret->procs = 0;
}

static char *check_oops(var * v, int formal_id)
{
	switch (v ? v->what : V_VOID) {
	case V_VOID:
		return "void";
		break;
	case V_STRING:
		if (formal_id != ID_DEF_STRING)
			return "string";
		break;
	case V_INT:
		if (formal_id != ID_DEF_INT)
			return "int";
		break;
	case V_PTR:
		if (formal_id != ID_DEF_PTR)
			return "ptr";
		break;
	default:
		abort();
	}
	return 0;
}

static int make_stack(vm_token *call, struct proc *proc, frame *f, frame *ret)
{
	byte *form, *real;
	vm_token formal;

	setup_frame(ret, f);
	
	for (real = call->left, form = proc->args; real && form;
	     form = formal.next) {
		var *v;
		char *oops;
		
		unpack(&formal, form);
		
		v = exec_nx(real, f, &real);

		if ((oops = check_oops(v, formal.id))) {
			croak(call, "%s result for arg '%s' in call to '%s'",
			     oops, formal.u.name, proc->name);
			free_var(v);
			return -1;
		}
		xfree(v->name);
		v->name = xstrdup(formal.u.name);
		drop_on_stack(v, ret);
	}

	if (form) {
		croak(call, "some arguments remained in call to '%s' "
			"(first: '%s')",
		      proc->name, formal.u.name);
		return -1;
	}
	
	if (real) {
		croak(call, "too many arguments in call to '%s'", proc->name);
		return -1;
	}
	return 0;	
}

var *exec_loop(byte *t, frame * f)
{
     	var *v = 0;
	byte *last;
	
	while (t && f->break_lev == 0 && f->continue_lev == 0
	       && f->return_lev == 0) {
	     	last = t;
		free_var(v);
		v = exec_nx(t, f, &t);

		if (f->parent == 0 && 
		    (f->continue_lev || f->break_lev || f->return_lev)) {
		    	vm_token x;
		    	
		    	unpack(&x, last);
			croak(&x, "cought %s at the top level",
			      f->continue_lev ? "continue" 
				       	      : (f->break_lev ? "break"
							      : "return"));
			f->continue_lev = f->break_lev = f->return_lev = 0;
		}
	}
	
	return v;
}

static var *do_call(vm_token *call, frame *f)
{
	frame local;
	struct proc *proc;
	u32 h;
	var *v;

	h = call->hash ? (u32)call->hash : xhash(call->u.name);
	
	if ((proc = 
		(struct proc*)hash_find_ex2(f->global->procs, 
							call->u.name, h)) == 0)
		proc = find_cproc(call->u.name, h);

	if (proc == 0) {
		croak(call, "call to undefined proc '%s'", call->u.name);
		return 0;
	}
		
	if (make_stack(call, proc, f, &local)) {
		kill_frame(&local);
		return 0;
	}

	local.active_token = call;
	
	if (proc->cproc)
		v = exec_cproc(proc, &local);
	else 
		v = exec_loop(proc->body, &local);

	kill_frame(&local);
	return v;
}

static var *do_dowhile(vm_token * t, frame * f)
{
	var *v = 0;
	
	do {
		free_var(v);
		v = exec_loop(t->left, f);

		if (f->continue_lev)
			f->continue_lev--;
			
		if (f->break_lev) {
			f->break_lev--;
			break;
		}
	} while (test_cond(t->u.top, f));
	return v;
}

static var *do_while(vm_token * t, frame * f)
{
	var *v = 0;
	
	while (test_cond(t->u.top, f)) {
		free_var(v);
		v = exec_loop(t->left, f);

		if (f->continue_lev)
			f->continue_lev--;
			
		if (f->break_lev) {
			f->break_lev--;
			break;
		}
	}
	return v;
}

static var *do_if(vm_token * t, frame * f)
{
	if (test_cond(t->u.top, f))
		return exec_loop(t->left, f);
	else
		return exec_loop(t->right, f);
}

static var *do_break(vm_token * t, frame * f, const char *name)
{
	var *v = 0;
	
	if (f->parent == 0) {
		croak(t, "cannot do %s at the top level", name);
		return 0;
	}

	if (t->left)
		v = exec(t->left, f);
		
	if (v == 0)
		v = new_iv(0, 1);

	if (*name != 'r' && v->what != V_INT) {
		croak(t, "%s argument is not a number", name);
		free_var(v);
		v = new_iv(0, 1);
	}
	
	if (*name == 'r')
		f->return_lev++;
	else if (*name == 'b')
		f->break_lev += v->var_int;
	else
		f->continue_lev += v->var_int;
	
	return v;
}

static var *do_get_var(vm_token *t, frame *f, int lval)
{
	var *v;
	u32 h;
	
	h = t->hash ? (u32)t->hash : xhash(t->u.name);
	
	if (t->u.name[0] == '$' 
	    && (v = get_magic_var(t, f)) != 0) {
		if (lval) {
			croak(t, "magic value (%s) cannot be used "
				"as a l-value", t->u.name);
			free_var(v);
			return 0;
		} else
			return v;
	} else
		v = get_var_h(t->u.name, f, h);
	if (v == 0)
		croak(t, "use of undefined variable '%s'", t->u.name);
	return lval ? v : copy(v);
}

static var *define_var(vm_token * t, frame * f, int type)
{
	var *v, *r = 0;
	
	if (get_var(t->u.name, f)) {
		croak(t, "redefinition of variable '%s'", t->u.name);
		return 0;
	}
	
	if (t->left)
		r = exec(t->left, f);
	
	switch (type) {
	case V_INT:
		v = new_iv(t->u.name, r ? r->var_int : 0);
		break;
	case V_STRING:
		v = new_sv(t->u.name, r ? r->var_string : "");
		break;
	case V_PTR:
		v = new_pv(t->u.name, r ? r->var_ptr : 0);
		break;
	default:
		abort();
	}
	free_var(r);
	drop_on_stack(v, f);
	return 0;
}

static var *do_array(vm_token *t, frame *f, int lval, var *v)
{
	var *i;
	int idx;
	
	i = exec(t->left, f);
	if (i == 0 || i->what != V_INT) {
		free_var(i);
		croak(t, "non-int (%s) type of array subscript", type_name(i));
		return 0;
	}
	idx = i->var_int;
	free_var(i);
	if (idx < 0 || idx >= (int)strlen(v->var_string))
		if (lval) {
			croak(t, "out of range lvalue array subscript "
				"(%d, max %d)", idx, strlen(v->var_string) - 1);
			return 0;
		} else
			return new_iv(0, 0);
	else
		if (lval)
			return new_adv(v->var_string + idx);
		else
			return new_iv(0, v->var_string[idx]);
}

static var *do_hash(vm_token * t, frame * f, int lval)
{
	var *v;

	v = get_var(t->u.name, f);
	if (v) {
		if (v->what == V_STRING) {
			return do_array(t, f, lval, v);
		} else {
			croak(t, "array subscript of non-string");
			return 0;
		}
	}
	return lval ? hash_entry(t, f) : copy(hash_entry(t, f));
}

var *exec_lval(byte *bc, frame *f)
{
	vm_token t;
	
	if (bc == 0)
		return 0;
	unpack(&t, bc);
	
	if (t.next)
		abort();
		
	switch (t.id) {
	default:
		croak(&t, "lvalue expected");
		return 0;
	case ID_VAR:
		return do_get_var(&t, f, 1);
	case ID_EXPR1:
		return do_expr1(&t, f, 1);
	case ID_EXPR2:
		return do_expr2(&t, f, 1);
	case ID_HASH:
		return do_hash(&t, f, 1);
	}
}

var *exec_nx(byte *bc, frame *f, byte **next)
{
	vm_token t;
	
	if (bc == 0)
		return 0;
	unpack(&t, bc);
	if (next)
		*next = t.next;
	else if (t.next)
		abort();
		
	switch (t.id) {
	case ID_ERROR:	
		abort();
	case ID_INT:
		return new_iv(0, t.u.number);
	case ID_STRING:
		return new_sv(0, t.u.name);
	case ID_VAR:
		return do_get_var(&t, f, 0);
	case ID_PROC:
		abort();
	case ID_CALL:
		return do_call(&t, f);
	case ID_DEF_PTR:
		return define_var(&t, f, V_PTR);
	case ID_DEF_STRING:
		return define_var(&t, f, V_STRING);
	case ID_DEF_INT:
		return define_var(&t, f, V_INT);
	case ID_DEF_SHASH:
		hash_add_one(f, t.u.name, V_STRING);
		return 0;
	case ID_DEF_PHASH:
		hash_add_one(f, t.u.name, V_PTR);
		return 0;
	case ID_DEF_IHASH:
		hash_add_one(f, t.u.name, V_INT);
		return 0;
	case ID_IF:
		return do_if(&t, f);
	case ID_DO_WHILE:
		return do_dowhile(&t, f);
	case ID_WHILE:
		return do_while(&t, f);
	case ID_WALK:
		return hash_walk(&t, f);
	case ID_BREAK:
		return do_break(&t, f, "break");
	case ID_CONTINUE:
		return do_break(&t, f, "continue");
	case ID_RETURN:
		return do_break(&t, f, "return");
	case ID_EXPR1:
		return do_expr1(&t, f, 0);
	case ID_EXPR2:
		return do_expr2(&t, f, 0);
	case ID_HASH:
		return do_hash(&t, f, 0);
	default:
		abort();
	}
}

var *exec(byte *bc, frame *f)
{
	return exec_nx(bc, f, 0);
}
